import math

import numpy as np
from Finch.layers import Populate, CapPopulation, SortByFitness
from Finch.environments import Sequential
from .layers import *
from .genepool import *
import pickle
from tqdm import tqdm


class TrieNode:
    def __init__(self):
        self.children = {}
        self.token_index = -1  # -1 indicates no token ends here
        self.is_end_of_token = False  # Indicates if a complete token ends at this node

class Trie:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, token, index):
        node = self.root
        for char in token:
            if char not in node.children:
                node.children[char] = TrieNode()
            node = node.children[char]
        node.token_index = index
        node.is_end_of_token = True  # Mark the end of a complete token

    def search(self, text):
        tokens = []
        node = self.root
        start_index = 0  # Start index of the current token being processed

        while start_index < len(text):
            node = self.root  # Reset node to root for each new starting character
            longest_token_index = -1  # Initialize with -1 indicating no token found yet
            longest_token_length = 0  # Length of the longest token found

            for i in range(start_index, len(text)):
                char = text[i]
                if char in node.children:
                    node = node.children[char]
                    if node.is_end_of_token:
                        longest_token_index = node.token_index  # Update if a longer token is found
                        longest_token_length = i - start_index + 1
                else:
                    break  # Break the loop if current character is not in children

            if longest_token_index != -1:
                tokens.append(longest_token_index)  # Append the longest token index found
                start_index += longest_token_length  # Move start index to the end of the longest token
            else:
                start_index += 1  # Move to the next character if no token was found

        return tokens


# todo identify gaps based on the halving rule
class GeneticTokenizer:
    def __init__(self, threshold=.001, min_range=2, max_range=6, max_population=11, start_population=10, mutate_amount=5, families=2, step_epochs: int = 1, existing_tokens: list = []):
        self.fitness_results = {}  # for speed boost
        self.tokens = existing_tokens
        self.step_epochs = step_epochs
        self.min_range = min_range
        self.max_range = max_range
        self.last_iteration = 0
        self.max_population = max_population
        self.start_population = start_population
        self.threshold = threshold
        self.mutate_amount = mutate_amount
        self.families = families
        self.trie = Trie()
        for i, token in enumerate(self.tokens):
            self.trie.insert(token, i)  # Populate Trie with existing tokens


    def evolve(self, dataset):
        total = len(dataset)
        with tqdm(total=total, desc="Evolving tokenizer") as pbar:
            for text in dataset:
                self.step(text)
                pbar.update(1)

    def step(self, text: str):
        pool = RangePool(min_range=self.min_range, max_range=self.max_range, source_text=text)

        max_population = self.max_population
        start_population = self.start_population

        # Create the environment
        environment = Sequential(
            layers=[
                Populate(gene_pool=pool, population=start_population),
                MutateToken(individual_selection=self.mutate_amount),
                ParentToken(families=self.families, gene_pool=pool),
                SortByFitness(),
                CapPopulation(max_population)
            ]
        )
        environment.compile(self.fitness, verbose_every=False)
        environment.iteration = self.last_iteration
        environment.evolve(self.step_epochs)
        self.last_iteration = environment.iteration

    def fitness(self, individual: RangeToken):
        token = individual.token
        if token in self.fitness_results:
            return self.fitness_results[token]

        source_text = individual.source
        count = source_text.count(token)
        percent = count / individual.length

        if percent > self.threshold:
            if token not in self.tokens:
                self.tokens.append(token)
                # Update the trie with the new token
                self.trie.insert(token, len(self.tokens) - 1)
        score = len(token) * percent
        self.fitness_results.update({token: score})

        return score

    def tokenize(self, text):
        indices = self.trie.search(text)
        return indices

    def detokenize(self, indices):
        """
        Detokenize the given list of indices into the original text.
        """
        return '|'.join(self.tokens[i] for i in indices)

    def interface(self):
        while 1:
            toks = self.tokenize(input("tokens: "))
            print("tokens: ", toks)
            print("detokens: ", self.detokenize(toks))

    def save(self, filename):
        """
        Save the state of the GeneticTokenizer object to a file, including the Trie.
        """
        with open(filename + ".gentok", 'wb') as f:
            pickle.dump({
                'fitness_results': self.fitness_results,
                'tokens': self.tokens,
                'step_epochs': self.step_epochs,
                'last_iteration': self.last_iteration,
                'trie': self.trie,  # Save the Trie structure
            }, f)

    def load(self, filename):
        """
        Load the state of the GeneticTokenizer object from a file, including the Trie.
        """
        with open(filename + ".gentok", 'rb') as f:
            data = pickle.load(f)

        self.tokens = data['tokens']
        self.fitness_results = data['fitness_results']
        self.step_epochs = data['step_epochs']
        self.last_iteration = data['last_iteration']
        self.trie = data['trie']  # Load the Trie structure



import random
from .layers import RangeToken
from Finch import genepools, layers


class RangePool(genepools.GenePool):
    def __init__(self, min_range, max_range, source_text: str):
        super().__init__(len(source_text), 'cpu')
        self.source_text = source_text
        self.min_range = min_range
        self.max_range = max_range

    def generate_individual(self):
        index = random.randint(0, self.length - self.max_range)

        _range = random.randint(self.min_range, self.max_range)
        return RangeToken(start=index, end=index + _range, source=self.source_text, gene_pool=self)



import random
from Finch import genepools, layers
from Finch.genetics import Individual
from typing import Union


class RangeToken(Individual):
    def __init__(self, gene_pool: genepools.GenePool, start: int, end: int, source: str):
        super().__init__(gene_pool=gene_pool, genes=None, fitness=1.0)
        self.start = start
        self.end = end
        self.source = source
        self.token = self.source[self.start:self.end]
        self.age = 0
        self.length = len(self.source)
        self.right_frozen = False
        self.left_frozen = False
        try:
            while self.token[0] == " ":
                self.start += 1
                self.token = self.token[1:-1]
                self.left_frozen = True
        except IndexError:
            pass
    def copy(self):
        return RangeToken(self.gene_pool, self.start, self.end, self.source)
    def mutate(self):
        choice = random.choice(['shift', 'length'])
        # if self.age > 10:
        #     self.redo()

        self.age += 1
        if choice == 'shift':
            if self.end < self.length + 1:
                shift = random.choice([1, -1])
                if not self.left_frozen:
                    self.start += shift
                self.end += shift
        if choice == "length":
            if self.end < self.length + 1:
                position = random.choice(['start', 'end'])
                if position == 'start' and not self.left_frozen:
                    self.start += random.choice([-1, 1])
                if position == 'end' or self.left_frozen:
                    self.end += random.choice([-1, 1])

        if self.end > self.start:
            self.end, self.start = self.start, self.end

        self.token = self.source[self.start:self.end]
        try:
            if self.token[0] == " ":
                self.start += 1
                self.token = self.token[1:-1]
                self.left_frozen = True
        except IndexError:
            pass
        # if self.token[-1] == " ":
        #     self.right_frozen True
    def redo(self):
        new = self.gene_pool.generate_individual()
        self.start = new.start
        self.end = new.end
        self.token = new.token
        self.left_frozen = new.left_frozen
        self.right_frozen = new.left_frozen
        self.age = 0

        try:
            while self.token[0] == " ":
                self.start += 1
                self.token = self.token[1:-1]
                self.left_frozen = True
        except IndexError:
            pass





def create_rangetoken(token: str, source_text: str, gene_pool: genepools.GenePool) -> Union[RangeToken, None]:
    try:
        start = source_text.index(token)
        end = start + len(token)
        token = RangeToken(gene_pool=gene_pool, start=start, end=end, source=source_text)
        return token
    except ValueError:
        return None
class MutateToken(layers.layer.Mutate):
    def __init__(self, individual_selection, refit: bool = True):
        super().__init__(individual_selection=individual_selection, refit=refit, gene_selection=None)
        self.selection = individual_selection

    def mutate_one(self, individual, environment):
        individual.mutate()


class ParentToken(layers.layer.Parent):
    def __init__(self, families: int, gene_pool: genepools.GenePool, refit=True):
        super().__init__(families=families, children=2, refit=refit, track_genealogies=True)
        self.gene_pool = gene_pool

    def parent(self, parent1: RangeToken, parent2: RangeToken, environment) -> list:
        a_token = parent1.token
        b_token = parent2.token

        a_child = a_token + b_token
        b_child = b_token + a_token

        a_child = create_rangetoken(a_child, source_text=parent1.source, gene_pool=self.gene_pool)
        b_child = create_rangetoken(b_child, source_text=parent2.source, gene_pool=self.gene_pool)

        return [child for child in [a_child, b_child] if child]



from setuptools import setup, find_packages

with open("genetok/README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("genetok/requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.readlines()

setup(
    name="genetok",
    version="0.1.0",
    author="Daniel Losey",
    author_email="danieljlosey@gmail.com",
    description="A Genetic Algorithm Tokenizer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dadukhankevin/genetok",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=requirements,
)
